open Base
open Stdio
open Expect_test_common.Std
open Expect_test_matcher.Std

module Test_result = Ppx_inline_test_lib.Runtime.Test_result
module Collector_test_outcome = Expect_test_collector.Test_outcome

type group =
  { filename      : File.Name.t
  ; file_contents : string
  ; tests         : Matcher.Test_outcome.t Map.M(File.Location).t
  }

let convert_collector_test ~filename (test : Collector_test_outcome.t)
  : File.Location.t * Matcher.Test_outcome.t =
  let saved_output =
    match Map.of_alist (module File.Location) test.saved_output with
    | `Ok x -> x
    | `Duplicate_key loc ->
      Printf.ksprintf failwith !"Output collector at %{File.Name}:%d ran more than once"
        filename loc.line_number ()
  in
  let expectations =
    List.map test.expectations ~f:(fun (expect : Expectation.Raw.t) ->
      (expect.extid_location,
       Expectation.map_pretty expect ~f:Lexer.parse_pretty)
    )
    |> Map.of_alist_exn (module File.Location)
  in
  (test.location,
   { expectations
   ; saved_output
   ; trailing_output = test.trailing_output
   })
;;

let create_group (filename, tests) =
  let module D = File.Digest in
  let expected_digest =
    match
      List.map tests ~f:(fun (t : Collector_test_outcome.t) -> t.file_digest)
      |> List.dedup ~compare:D.compare
    with
    | [digest] -> digest
    | [] -> assert false
    | digests ->
      Printf.ksprintf failwith
        !"Expect tests make inconsistent assumption about file \"%{File.Name}\" \
          %{sexp:D.t list}"
        filename digests
  in
  let file_contents =
    In_channel.read_all (File.Name.relative_to ~dir:(File.initial_dir ()) filename)
  in
  let current_digest =
    Caml.Digest.string file_contents |> Caml.Digest.to_hex |> D.of_string
  in
  if D.compare expected_digest current_digest <> 0 then
    Printf.ksprintf failwith
      !"File \"%{File.Name}\" changed, you need rebuild inline_test_runner \
        to be able to run expect tests \
        (expected digest: %{D}, current digest: %{D})"
      filename expected_digest current_digest;
  let tests =
    List.map tests ~f:(convert_collector_test ~filename)
    |> Map.of_alist_exn (module File.Location)
  in
  { filename
  ; file_contents
  ; tests = tests
  }
;;

let convert_collector_tests tests : group list =
  List.map tests ~f:(fun (test : Collector_test_outcome.t) ->
    (test.location.filename, test))
  |> Map.of_alist_multi (module File.Name)
  |> Map.to_alist
  |> List.map ~f:create_group
;;

let process_group ~use_color ~in_place ~diff_command { filename; file_contents; tests }
  : Test_result.t =
  let bad_outcomes =
    Map.fold tests ~init:[] ~f:(fun ~key:location ~data:test acc ->
      match Matcher.evaluate_test ~file_contents ~location test with
      | Match -> acc
      | Correction c -> c :: acc)
    |> List.rev
  in
  let filename = File.Name.relative_to ~dir:(File.initial_dir ()) filename in
  let dot_corrected = filename ^ ".corrected" in
  let remove_dot_corrected () =
    if Caml.Sys.file_exists dot_corrected then Caml.Sys.remove dot_corrected
  in
  match bad_outcomes with
  | [] ->
    remove_dot_corrected ();
    Success
  | _ ->
    Matcher.write_corrected bad_outcomes
      ~file:(if in_place then filename else dot_corrected)
      ~file_contents ~mode:Inline_expect_test;
    if in_place then begin
      remove_dot_corrected ();
      Success
    end else begin
      Print_diff.print ~file1:filename ~file2:dot_corrected ~use_color ?diff_command ();
      Failure
    end
;;

let evaluate_tests ~use_color ~in_place ~diff_command =
  convert_collector_tests (Expect_test_collector.tests_run ())
  |> List.map ~f:(process_group ~use_color ~in_place ~diff_command)
  |> Test_result.combine_all
;;

let () =
  Ppx_inline_test_lib.Runtime.add_evaluator ~f:(fun () ->
    evaluate_tests
      ~use_color:Ppx_inline_test_lib.Runtime.use_color
      ~in_place:Ppx_inline_test_lib.Runtime.in_place
      ~diff_command:Ppx_inline_test_lib.Runtime.diff_command)
;;
